import importlib


class SelectorExtractor(object):

    @staticmethod
    def get_list_data(elements=None):
        data_cleaned = []
        data = elements.extract()
        for i, datum in enumerate(data):
            if datum:
                data_cleaned.append(datum.strip())
        return data_cleaned

    @staticmethod
    def get_single_data(elements=None):
        data = elements.extract_first()
        if data:
            return data.strip()
        return data


def transform_data(data=None, data_type=None):
    fields = importlib.import_module("invana_bot.fields")
    Klass = getattr(fields, data_type)
    data = Klass(data=data).transform()
    return data


def clean_data(elements=None, selector=None):
    """

    This is where are the extracted data will be cleaned up and applied functions and data types as needed.

    :param elements:
    :param selector:
    :return:
    """
    data_type = selector.get("data_type", "RawField")

    if data_type.startswith("List"):
        multiple = True
    else:
        multiple = False

    data_extractor = SelectorExtractor()
    if multiple is True:
        extracted_data = data_extractor.get_list_data(elements=elements)
    else:
        extracted_data = data_extractor.get_single_data(elements=elements)
    data = transform_data(data=extracted_data, data_type=data_type)
    return data


def get_selector_element(html_element, selector):
    if selector.get('selector_attribute') in ['text']:
        if selector.get('selector_type') == 'css':
            elements = html_element.css("{0}::{1}".format(selector.get('selector'),
                                                          selector.get('selector_attribute')))
            return clean_data(elements=elements, selector=selector)
        else:
            elements = html_element.xpath("{0}/{1}".format(selector.get('selector'),
                                                           selector.get('selector_attribute')))
            return clean_data(elements=elements, selector=selector)
    elif selector.get('selector_attribute') == 'html':
        if selector.get('selector_type') == 'css':
            elements = html_element.css(selector.get('selector'))
            return clean_data(elements=elements, selector=selector)
        else:
            elements = html_element.xpath("{0}/{1}".format(selector.get('selector'),
                                                           selector.get('selector_attribute')))
            return clean_data(elements=elements, selector=selector)
    else:
        if selector.get('selector_type') == 'css':
            elements = html_element.css(selector.get('selector')) \
                .xpath("@{0}".format(selector.get('selector_attribute')))
            return clean_data(elements=elements, selector=selector)
        else:
            elements = html_element.xpath("{0}/{1}".format(selector.get('selector'),
                                                           selector.get('selector_attribute')))
            return clean_data(elements=elements, selector=selector)



from scrapy.spiders import CrawlSpider
from scrapy.http import Request
from invana_bot.utils.spiders import get_spider_from_list
import os
from invana_bot.traversals.generic import GenericLinkExtractor
import scrapy


class WebCrawlerBase(CrawlSpider):
    """

    TODO - document why we need this as base method.

    """

    spider_id = None
    spider_config = None
    spider_data_storage = None
    manifest = {}

    def post_parse(self, response=None):
        pass

    def parse_error(self, failure):
        pass

    def start_requests(self):
        for url in self.start_urls:
            yield scrapy.Request(
                url,
                callback=self.parse,
                errback=self.parse_error,
                dont_filter=True,
                meta={
                    "current_request_traversal_page_count": 0,
                    "spider_config": self.spider_config,
                    "manifest": self.manifest
                }
            )

    def get_spider_config(self, response=None):
        if response.meta.get("spider_config"):
            return response.meta.get("spider_config")
        else:
            return self.spider_config

    @staticmethod
    def get_default_storage(settings=None, spider_config=None):
        data_storages = settings.get("DATA_STORAGES", [])
        default_storage = None
        spider_storage_id = "default"
        for data_storage in data_storages:
            __storage_id = data_storage.get("storage_id") or data_storage.get("STORAGE_ID")
            if __storage_id == spider_storage_id:
                return data_storage
        return default_storage

    @staticmethod
    def prepare_data_for_yield(data=None, collection_name=None, storage_id="default"):
        return {
            "_data_storage_id": storage_id,
            "_data_storage_collection_name": collection_name,
            "_data": data
        }

    def _build_request(self, rule, link):
        headers = {}
        user_agent_header = os.environ.get("WCP_REQUEST_HEADERS_USER_AGENT")
        if user_agent_header:
            headers = {"User-Agent": user_agent_header}
        r = Request(url=link.url, headers=headers, callback=self._response_downloaded)
        r.meta.update(rule=rule, link_text=link.text)
        return r

    @staticmethod
    def is_this_request_from_same_traversal(response, traversal):
        """
        This mean the current request came from this  traversal,
        so we can put max pages condition on this, otherwise for different
        traversals of different spiders, adding max_page doest make sense.
        """
        traversal_id = traversal['traversal_id']
        current_request_traversal_id = response.meta.get('current_request_traversal_id', None)
        return current_request_traversal_id == traversal_id

    def make_traversal_requests(self, to_traverse_links_list=None):
        traversal_requests = []
        for to_traverse_link in to_traverse_links_list:
            traversal_requests.append(scrapy.Request(
                to_traverse_link.get("link"),
                callback=self.parse,
                errback=self.parse_error,
                meta=to_traverse_link.get("meta", {})
            ))
        return traversal_requests

    @staticmethod
    def run_traversal(response=None, traversal=None, **kwargs):

        selector_type = traversal.get("selector_type", "css")
        kwargs = {}
        if selector_type == "css":
            kwargs['restrict_css'] = (traversal.get("selector_value"),)
        elif selector_type == "xpath":
            kwargs['restrict_xpaths'] = (traversal.get("selector_value"),)

        if traversal.get("allow_domains", []) == ["*"]:
            kwargs['allow_domains'] = ()
        else:
            kwargs['allow_domains'] = traversal.get("allow_domains", [])
        return GenericLinkExtractor(**kwargs).extract_links(response=response)

    def run_traversals(self, spider_config=None, response=None):
        """
        if spider_traversal_id is None, it means this response originated from the
        request raised by the start urls.

        If it is Not None, the request/response is raised some traversal strategy.
        """
        current_request_traversal_id = response.meta.get('current_request_traversal_id', None)
        current_request_traversal_page_count = response.meta.get('current_request_traversal_page_count', 0)

        """
        Note on current_request_spider_id:
        This can never be none, including the ones that are started by start_urls .
        """
        traversal_data = {}
        to_traverse_links_list = []
        spider_traversals = spider_config.get('traversals', [])
        spiders = response.meta.get("manifest", {}).get("spiders")

        for traversal in spider_traversals:
            next_spider_id = traversal['next_spider_id']
            next_spider = get_spider_from_list(spider_id=next_spider_id, spiders=spiders)

            traversal['allow_domains'] = next_spider.get("allowed_domains", [])
            traversal_id = traversal['traversal_id']
            traversal_max_pages = traversal.get('max_pages', 1)

            traversal_links = []
            is_this_request_from_same_traversal = self.is_this_request_from_same_traversal(response, traversal)
            # print("is_this_request_from_same_traversal", is_this_request_from_same_traversal)
            # print("current_request_traversal_page_count", current_request_traversal_page_count)
            # print("traversal_max_pages", traversal_max_pages)
            # print(" current_request_traversal_page_count < traversal_max_pages",
            # current_request_traversal_page_count < traversal_max_pages)
            shall_traverse = False

            if current_request_traversal_id is None:
                """
                start urls will not have this traversal_id set, so we should allow then to traverse
                """
                shall_traverse = True

            elif is_this_request_from_same_traversal and current_request_traversal_page_count < traversal_max_pages:
                """
                This block will be valid for the traversals from same spider_id, ie., pagination of a spider 
                """

                shall_traverse = True

            elif is_this_request_from_same_traversal:
                """
                """
                shall_traverse = True

            elif is_this_request_from_same_traversal is False and current_request_traversal_page_count < traversal_max_pages:
                """
                This for the spider_a traversing to spider_b, this is not pagination, but trsversing between 
                spiders.
                """
                shall_traverse = True
            # print("shall_traverse: {}".format(traversal_id), shall_traverse)
            if shall_traverse:
                traversal_links = self.run_traversal(response=response, traversal=traversal)
                traversal_data[traversal_id] = {"traversal_urls": traversal_links}
                """
                Then validate for max_pages logic if traversal_id's traversal has any!.
                This is where the further traversal for this traversal_id  is decided 
                """
                max_pages = traversal.get("max_pages", 1)
                for link in traversal_links:

                    """
                    we are already incrementing, the last number, so using <= might make it 6 pages when 
                    max_pages is 5 
                    """
                    if current_request_traversal_page_count < max_pages:
                        # print("=======current_request_traversal_page_count", current_request_traversal_page_count)
                        # print("link", link)
                        # print("-----------------------------------")

                        # print("===>>>>meta", meta)
                        to_traverse_links_list.append(
                            {
                                "link": link,
                                "meta": {
                                    "spider_config": next_spider,
                                    "manifest": response.meta.get("manifest"),
                                    "current_request_traversal_id": traversal_id,
                                    "current_request_traversal_page_count": current_request_traversal_page_count,

                                }}
                        )

                    current_request_traversal_page_count += 1

            print("=================================================")
            print("{} traversal_links for traversal:".format(len(traversal_links), traversal_id))
            print("=================================================")
        return traversal_data, to_traverse_links_list



from invana_bot.extractors.base import ExtractorBase


class PythonBasedExtractor(ExtractorBase):
    """


    def extractor_fn(response=response):
        html_content = response.title

        return {"data": {}, "d__d": []}


    """

    def run(self):
        print("-======response", type(self.response), self.extractor)
        extractor_fn = self.extractor.get("extractor_fn")
        print ("extractor_fn=", extractor_fn)
        # exit()
        data = {}
        if extractor_fn:
            data[self.extractor_id] = extractor_fn(response=self.response)
        return data



from pymongo import MongoClient


class Executor:
    def connect(self):
        raise NotImplementedError

    def read(self):
        raise NotImplementedError

    def disconnect(self):
        raise NotImplementedError



from __future__ import print_function
import logging
from scrapy.responsetypes import responsetypes
from scrapy.utils.request import request_fingerprint
from scrapy.utils.python import to_bytes
import pymongo
from scrapy.http.headers import Headers
from invana_bot.utils.url import get_domain

logger = logging.getLogger(__name__)


class MongoDBCacheStorage(object):
    """
    should set INVANA_BOT_SETTINGS in the default.py

    pipeline_settings = {
        'ITEM_PIPELINES': {'invana_bot.storages.mongodb.MongoDBPipeline': 1},
        'HTTPCACHE_STORAGE': "invana_bot.httpcache.mongodb.MongoDBCacheStorage",
    }

    mongodb_settings = {
        'INVANA_BOT_SETTINGS': {
            'HTTPCACHE_STORAGE_SETTINGS': {
                'CONNECTION_URI': "mongodb://127.0.0.1",
                'DATABASE_NAME': "spider_cache_db",
                'COLLECTION_NAME': "web_link",
                "EXPIRY_TIME": 3600
            },
            'ITEM_PIPELINES_SETTINGS': {
                'CONNECTION_URI': "mongodb://127.0.0.1",
                'DATABASE_NAME': "spider_data",
                'COLLECTION_NAME': "spider_feeds_data"
            }
        }
    }


    """

    def __init__(self, settings):
        self.CONNECTION_URI = settings.get('INVANA_BOT_SETTINGS', {}).get('HTTPCACHE_STORAGE_SETTINGS', {}).get(
            "CONNECTION_URI", None)
        self.database_name = settings.get('INVANA_BOT_SETTINGS', {}).get('HTTPCACHE_STORAGE_SETTINGS', {}).get(
            "DATABASE_NAME", None)
        self.cache_expiry_time = settings.get('INVANA_BOT_SETTINGS', {}).get('HTTPCACHE_STORAGE_SETTINGS', {}).get(
            "EXPIRY_TIME", None)
        self.collection_name = settings.get('INVANA_BOT_SETTINGS', {}).get('HTTPCACHE_STORAGE_SETTINGS', {}).get(
            "COLLECTION_NAME", None)
        self.db_client = pymongo.MongoClient(self.CONNECTION_URI)
        self.db = self.db_client[self.database_name]

    def open_spider(self, spider):
        logger.debug("Using mongodb cache storage with database name %(database)s" % {'database': self.database_name},
                     extra={'spider': spider})

    def close_spider(self, spider):
        pass

    def retrieve_response(self, spider, request):
        data = self._read_data(spider, request)
        if data is None:
            return  # not cached
        url = data['url']
        status = data['status']
        headers = Headers(data['headers'])
        body = data['html']
        respcls = responsetypes.from_args(headers=headers, url=url)
        response = respcls(url=url, headers=headers, status=status, body=body)
        return response

    def _clean_headers(self, obj):
        cleaned_object = {}
        for k, v in obj.items():
            cleaned_object[k.decode('utf-8')] = v[0].decode('utf-8')
        return cleaned_object

    def store_response(self, spider, request, response):
        data = {
            'status': response.status,
            'domain': get_domain(response.url),
            'url': response.url,
            'headers': self._clean_headers(response.headers),
            'html': response.body,
        }
        self.db[self.collection_name].insert_one(data)

    def _read_data(self, spider, request):
        return self.db[self.collection_name].find_one({'url': request.url})

    def _request_key(self, request):
        return to_bytes(request_fingerprint(request))



from .single import StringField, IntField, FloatField, DictField, RawField


class ListFieldTransformerBase(object):
    transformer_class = None

    def __init__(self, data=None):
        self.data = data
        self.validate_transformer()

    def validate_transformer(self):
        if self.transformer_class is None:
            raise NotImplementedError("transformer_class should be assigned for List type of FieldTransformer")

    def try_or_none(self):

        result_data_list = []
        try:
            result_data = self.transformer_class(data=self.data).transform()
            result_data_list.append(result_data)
        except Exception as e:
            print(e)
            result_data = self.data
            result_data_list.append(result_data)
        return result_data_list

    def transform(self):
        return self.try_or_none()


class ListStringField(ListFieldTransformerBase):
    transformer_class = StringField


class ListIntField(ListFieldTransformerBase):
    transformer_class = IntField


class ListFloatField(ListFieldTransformerBase):
    transformer_class = FloatField


class ListDictField(ListFieldTransformerBase):
    transformer_class = DictField


class ListRawField(ListFieldTransformerBase):
    transformer_class = RawField



from datetime import datetime
from scrapy.exceptions import DropItem
from elasticsearch import Elasticsearch
from pymongo import MongoClient
from invana_bot.utils.storage import generate_random_id
from slugify import slugify


class InvanaDataPipeline(object):

    @staticmethod
    def create_mongodb_connection(data_storage=None):
        db_client = MongoClient(data_storage.get("connection_uri"))
        db_conn = db_client[data_storage.get("database_name")]
        return db_conn

    @staticmethod
    def create_elasticsearch_connection(data_storage=None):

        db_conn = Elasticsearch(
            [{'host': data_storage.get("connection_uri"), 'port': 9200}])
        return db_conn

    def __init__(self, data_storages=None):

        self.data_storage_conns = {}
        for data_storage in data_storages:
            storage_type = data_storage.get("storage_type", "mongodb")
            storage_id = data_storage.get("storage_id", "default")
            if storage_type == "mongodb":
                self.data_storage_conns[storage_id] = {
                    "_connection": self.create_mongodb_connection(data_storage=data_storage),
                    "_data_storage": data_storage
                }
            elif storage_type == "elasticsearch":
                self.data_storage_conns[storage_id] = {
                    "_connection": self.create_elasticsearch_connection(data_storage=data_storage),
                    "_data_storage": data_storage
                }
            else:
                raise NotImplementedError("yet to implement '{}' pipeline".format(storage_type))

    @classmethod
    def from_crawler(cls, spider):
        return cls(
            data_storages=spider.settings['DATA_STORAGES']
        )

    def process_item(self, item, spider):
        """

        :param item:  {
            "_data_storage_id": "default",
            "_data_storage_collection_name": "mycollection",
            "_data": <data to save>
        }
        :param spider:
        :return:
        """
        """
        comes handy during the custom data saving.

        data_storage_id = item.get("_data_storage_id")

        """
        data_storage_id = "default"
        data_storage = self.data_storage_conns.get(data_storage_id)
        data_storage_conn = data_storage.get("_connection")
        data_storage_collection_name = data_storage.get("_data_storage", {}).get("collection_name")

        data = item.get("_data")
        if None in [data_storage, data_storage_conn]:
            raise DropItem(
                "Data storage with id: {} doesn't exist, so dropping the item {}".format(data_storage_id, item))

        data = dict(data)
        if "updated" not in data.keys():
            data['updated'] = datetime.now()
        data_storage_type = data_storage.get("storage_type") or "mongodb"
        if data_storage_type == "mongodb":
            data_storage_conn[data_storage_collection_name].insert(data)
        elif data_storage_type == "elasticsearch":

            unique_key = data_storage.get("_data_storage", {}).get("unique_key")
            if unique_key and unique_key in data.items():
                doc_id = slugify(data[unique_key])
            else:
                doc_id = generate_random_id()
            data_storage_conn.index(
                index=data_storage.get("_data_storage", {}).get("database_name"),
                doc_type=data_storage_collection_name,
                id=doc_id,
                body=data)
        else:
            raise NotImplementedError(
                "storage_type:{} not implemeneted, so extracted data is not saved.".format(data_storage_type))

        return item



from invana_bot.extractors.base import ExtractorBase


class ImagesExtractor(ExtractorBase):
    def run(self):
        data = {}
        images_data = []
        images_selector = self.response.xpath('//img/@src').extract()
        for selector in images_selector:
            images_data.append(selector)
        data[self.extractor_id] = images_data
        return data



from scrapy.linkextractors import LinkExtractor


class GenericLinkExtractor(object):
    """


    """

    def __init__(self,
                 restrict_xpaths=(),
                 restrict_css=(),
                 restrict_regex=(),
                 allow_domains=(),
                 link_extractor_cls=LinkExtractor, **kwargs):
        """

        :param restrict_xpaths: list of xpaths for links Extraction.
        :param restrict_css: list of xpath for links extraction
        :param restrict_regex: list of regex patterns
        :param link_extractor_cls: defaults to scrapy link extractor
        :param allow_domains: defaults to the allowed domains of spider
        """
        self.restrict_xpaths = restrict_xpaths
        self.restrict_css = restrict_css
        self.restrict_regex = restrict_regex
        self.allow_domains = allow_domains
        self.link_extractor_cls = link_extractor_cls

    def extract_links(self, response=None):
        all_links = self.link_extractor_cls(allow=self.restrict_xpaths,
                                            restrict_xpaths=self.restrict_xpaths,
                                            restrict_css=self.restrict_css,
                                            allow_domains=self.allow_domains
                                            ).extract_links(response=response)
        return [link.url for link in all_links]



def default_transformer(results):
    results_cleaned = results
    return results_cleaned


class InvanaBotTranformerBase(object):

    def __init__(self, cti_config=None,
                 transformer_name=None,
                 cit_id=None,
                 crawled_id=None,
                 job_id=None):

        if None in [transformer_name,
                    cit_id, crawled_id, job_id]:
            raise Exception("transformer_name, cit_id, crawled_id, job_id "
                            "should be given as input for InvanaBotTranformer ")
        self.cti_config = cti_config
        self.transformer_name = transformer_name
        self.cit_id = cit_id
        self.crawled_id = crawled_id
        self.job_id = job_id

    def transform(self):
        raise NotImplementedError("this method should be implemented")

    def filter_data(self):
        pass



# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import argparse
from invana_bot.jobs.cti import InvanaBotJobGenerator
from invana_bot.settings.default import DEFAULT_SETTINGS
from invana_bot.manifests.cti import CTIManifestManager
from invana_bot.spiders.web import InvanaBotSingleWebCrawler
from invana_bot.spiders.xml import GenericXMLFeedSpider
from invana_bot.spiders.api import GenericAPISpider
# from invana_bot.jobs.single import SingleCrawlJobGenerator
# from invana_bot.manifests.single import SingleCrawlerManifestManager


def invana_bot_run():
    spider_choices = (
        "web",
        "web-single",
        "xml",
        "api"
    )
    parser = argparse.ArgumentParser(description='InvanaBot - A web spider framework that can'
                                                 ' transform websites into datasets with Crawl, '
                                                 'Transform and Index workflow; just with the configuration.')

    parser.add_argument('--path', type=str, default='./', help='The path of the cti_manifest.yml')
    parser.add_argument('--type', type=str,
                        default='web',
                        required=True,
                        help='options are : {}'.format(",".join(spider_choices)),
                        choices=spider_choices)

    args = parser.parse_args()
    path = os.path.abspath(args.path)
    spider_type = args.type

    if spider_type == "web":
        spider_cls = InvanaBotSingleWebCrawler
    elif spider_type == "xml":
        spider_cls = GenericXMLFeedSpider
    elif spider_type == "api":
        spider_cls = GenericAPISpider

    else:
        raise Exception("There is no crawling strategy designed for spider type: '{}'".format(spider_type))

    manifest_manager = CTIManifestManager(
        manifest_path=path
    )

    manifest, errors = manifest_manager.get_manifest()
    # print("cti_manifest", cti_manifest)

    first_spider = manifest.get("spiders", [])[0]

    ignore_spider_keys = ["spider_id", "allowed_domains", "extractors", "traversals"]
    extra_arguments = {}
    for k, v in first_spider.items():
        if k not in ignore_spider_keys:
            extra_arguments[k] = v
    if len(errors) == 0:
        spider_job_generator = InvanaBotJobGenerator(
            settings=DEFAULT_SETTINGS,
        )
        context = manifest.get("context")
        job = spider_job_generator.create_job(
            manifest=manifest,
            context=context,
            spider_cls=spider_cls,
            extra_arguments=extra_arguments

        )
        spider_job_generator.start_job(job=job)
    else:
        print("==============================================================")
        print("ERROR : ETI Job Failing with the errors :: {}".format(
            manifest_manager.manifest_path,
            errors
        ))
        print("==============================================================")

        # manifest_manager = CTIManifestManager(
        #     manifest_path=path
        # )
        # cti_manifest, errors = manifest_manager.get_manifest()
        # if len(errors) == 0:
        #     spider_job_generator = InvanaBotJobGenerator(
        #         settings=DEFAULT_SETTINGS,
        #     )
        #     context = cti_manifest.get("context")
        #     job = spider_job_generator.create_job(
        #         cti_manifest=cti_manifest,
        #         context=context
        #     )
        #     spider_job_generator.start_job(job=job)
        # else:
        #     print("==============================================================")
        #     print("ERROR : ETI Job Failing with the errors :: {}".format(
        #         manifest_manager.manifest_path,
        #         errors
        #     ))
        #     print("==============================================================")
        #
    #
    # if spider_type == "web-single":
    #     manifest_manager = SingleCrawlerManifestManager(
    #         config_path=path
    #     )
    #     spider_manifest, errors = manifest_manager.get_manifest()
    #     if len(errors) == 0:
    #         spider_job_generator = SingleCrawlJobGenerator(
    #             settings=DEFAULT_SETTINGS
    #         )
    #         context = spider_manifest.get("context")
    #         job = spider_job_generator.create_job(
    #             spider_config=spider_manifest,
    #             context=context
    #         )
    #         spider_job_generator.start_job(job=job)
    #     else:
    #         print("==============================================================")
    #         print("ERROR : ETI Job Failing with the errors :: {}".format(
    #             manifest_manager.config_path,
    #             errors
    #         ))
    #         print("==============================================================")



from invana_bot.extractors.base import ExtractorBase
from invana_bot.utils.selectors import get_selector_element
from invana_bot.utils.url import get_urn, get_domain


class AllLinksExtractor(ExtractorBase):
    def run(self):
        data = {}
        extracted_data = []
        links = self.response.css("a").xpath("@href").extract()

        for link in links:
            if link and not link.startswith("#"):
                extracted_data.append(link)
        data[self.extractor_id] = extracted_data
        return data


class AllLinksAnalyticsExtractor(ExtractorBase):

    def run(self):
        data = {}
        extracted_data = AllLinksExtractor(
            response=self.response,
            extractor=self.extractor,
            extractor_id="all_links"
        ).run().get("all_links", {})

        links_data = {}
        for link in extracted_data:
            domain = get_domain(link)
            if domain in links_data.keys():
                links_data[domain].append(link)
            else:
                links_data[domain] = [link]

        data[self.extractor_id] = [{"domain": domain, "links": domain_links, "links_count": domain_links.__len__()} for
                                domain, domain_links in links_data.items()]

        return data


class SameDomainLinkExtractor(ExtractorBase):
    pass


class ForeignDomainLinkExtractor(ExtractorBase):
    pass

#
# class PaginationLinkExtractor(ExtractorBase):
#     def run(self):
#         data = {}
#         extracted_data = {}
#         for selector in self.extractor.get('data_selectors', []):
#             _d = get_selector_element(self.response, selector)
#             extracted_data[selector.get('selector_id')] = _d
#         data[self.extractor_id] = extracted_data
#         return data


class CustomLinkExtractor(ExtractorBase):
    pass



from __future__ import unicode_literals
from pymongo import MongoClient
from invana_bot.transformers.executors import Executor
from datetime import datetime


class WriteToMongoDB(Executor):
    def __init__(self, connection_string, db, collection, unique_key, docs=None, update=True):
        self._connection_string = connection_string
        self._db = db
        self._collection = collection
        self._docs = docs
        self._unique_key = unique_key

    def connect(self):
        self._client = MongoClient(self._connection_string)

    def get_object(self, doc=None):
        return self._client[self._db][self._collection].find_one({self._unique_key: doc.get(self._unique_key)})

    def write(self):

        if isinstance(self._client, MongoClient):
            for doc in self._docs:
                entry = self.get_object(doc=doc)
                if entry is None:
                    self._client[self._db][self._collection].insert(doc)
                else:
                    doc['updated'] = datetime.now()
                    self._client[self._db][self._collection].update_one({self._unique_key: doc.get(self._unique_key)},
                                                                        {"$set": doc})

    def disconnect(self):
        if isinstance(self._client, MongoClient):
            self._client.close()


class ReadFromMongo(Executor):
    def __init__(self, connection_string, db, collection, query=None, fields=None):
        self._connection_string = connection_string
        self._db = db
        self._collection = collection
        self._query = query
        self._fields = fields
        self._client = None

    def connect(self):
        self._client = MongoClient(self._connection_string)

    def read(self):
        if isinstance(self._client, MongoClient):
            for doc in self._client[self._db][self._collection].find(self._query, projection=self._fields):
                yield doc

    def disconnect(self):
        if isinstance(self._client, MongoClient):
            self._client.close()


class OTManager(object):
    """
        Object Transform Manager Renders the output into required form
        Arguments
        ---------
        ops - list of OTConf objects
    """

    def __init__(self, ops):
        self._ops = ops
        self.results = []

    def _sub_process(self, op, _object):
        if op.key_path in _object:
            op.cls(op.key_path, *op.args, **op.kwargs).process(_object)
        elif '.' in op.key_path:
            key_split = op.key_path.split('.')
            _op = op.clone()
            _op.key_path = '.'.join(key_split[1:])
            if isinstance(_object[key_split[0]], list):
                for _list_item in _object[key_split[0]]:
                    self._sub_process(_op, _list_item)
            else:
                self._sub_process(_op, _object[key_split[0]])
        return _object

    def process(self, executor):
        for _object in executor.read():
            for op in self._ops:
                self._sub_process(op, _object)
            self.results.append(_object)
        return self



from invana_bot.utils.spiders import get_spider_from_list
from invana_bot.utils.config import validate_cti_config
from .base import RunnerEngineBase
from .single import InvanaBotSingleCrawlerRunnerEngine


class InvanaBotRunnerEngine(RunnerEngineBase):
    """

    InvanaBotRunnerEngine


    """

    def __init__(self, job_id=None, spider_cls=None, manifest=None):
        self.job_id = job_id
        self.spider_cls = spider_cls
        self.manifest = manifest

    def crawl(self):
        errors = validate_cti_config(self.manifest)
        if len(errors) == 0:

            initial_spider_config = get_spider_from_list(
                spider_id=self.manifest['init_spider']['spider_id'],
                spiders=self.manifest['spiders']
            )
            initial_spider_config['start_urls'] = self.manifest['init_spider']['start_urls']
            spider_runner = InvanaBotSingleCrawlerRunnerEngine(
                job_id=self.job_id,
                spider_cls=self.spider_cls,
                spider_config=initial_spider_config,
                manifest=self.manifest,
            )
            job = spider_runner.run()

            return job, errors
        else:
            return None, errors



from __future__ import print_function
import logging
from scrapy.responsetypes import responsetypes
from scrapy.utils.request import request_fingerprint
from scrapy.utils.python import to_bytes
from scrapy.http.headers import Headers
from elasticsearch_dsl import DocType, Date, Integer, Text, connections
from datetime import datetime
from invana_bot.utils.url import get_urn, get_domain

logger = logging.getLogger(__name__)


class ESCacheStorage(object):
    """
    pipeline_settings = {
        'ITEM_PIPELINES': {'invana_bot.storages.elasticsearch.ElasticSearchPipeline': 1},
        'HTTPCACHE_STORAGE': "invana_bot.httpcache.elasticsearch.ESCacheStorage",
    }

    es_settings = {
        'INVANA_BOT_SETTINGS': {
            'HTTPCACHE_STORAGE_SETTINGS': {
                'CONNECTION_URI': "127.0.0.1",
                'DATABASE_NAME': "spider_cache_db",
                'COLLECTION_NAME': "web_link",
                "EXPIRY_TIME": 3600
            },
            'ITEM_PIPELINES_SETTINGS': {
                'CONNECTION_URI': "127.0.0.1",
                'DATABASE_NAME': "spider_data",
                'COLLECTION_NAME': "spider_feeds_data"
            }
        }
    }

    """

    def setup_collection(self):
        class WebLink(DocType):
            url = Text()
            html = Text()
            headers = Text()
            status = Integer()
            created = Date()

            class Meta:
                index = self.database_name
                doc_type = self.collection_name

        return WebLink

    def __init__(self, settings):

        self.CONNECTION_URI = settings.get('INVANA_BOT_SETTINGS', {}).get('HTTPCACHE_STORAGE_SETTINGS', {}).get(
            "CONNECTION_URI", None)
        self.database_name = settings.get('INVANA_BOT_SETTINGS', {}).get('HTTPCACHE_STORAGE_SETTINGS', {}).get(
            "DATABASE_NAME", None)
        self.cache_expiry_time = settings.get('INVANA_BOT_SETTINGS', {}).get('HTTPCACHE_STORAGE_SETTINGS', {}).get(
            "EXPIRY_TIME", None)
        self.collection_name = settings.get('INVANA_BOT_SETTINGS', {}).get('HTTPCACHE_STORAGE_SETTINGS', {}).get(
            "COLLECTION_NAME", None)
        connections.create_connection(hosts=[self.CONNECTION_URI])
        self.WebLink = self.setup_collection()
        self.WebLink.init()

    def open_spider(self, spider):
        logger.debug("Using elastic cache storage with index name %(database)s" % {'database': self.CONNECTION_URI},
                     extra={'spider': spider})

    def close_spider(self, spider):
        pass

    def get_headers(self, obj):
        """
        this will convert all the headers_Server, headers_Date
        into "header": {
            "Server": "",
            "Date": ""
        }

        :param obj:
        :return:
        """
        headers = {}
        for k, v in obj.items():
            if k.startswith("headers_"):
                headers[k.replace("headers_", "")] = v

        obj['headers'] = headers
        return obj

    def retrieve_response(self, spider, request):
        data = self._read_data(spider, request)

        if data is None:
            return  # not cached
        else:
            if data['status'] == 200 and data['html'] is None:
                return None

        data = self.get_headers(data)
        url = data['url']
        status = data['status']
        headers = Headers(data['headers'])
        body = bytes(data['html'], encoding="utf-8")
        respcls = responsetypes.from_args(headers=headers, url=url)
        response = respcls(url=url, headers=headers, status=status, body=body)
        return response

    def _clean_headers(self, obj):
        cleaned_object = {}
        for k, v in obj.items():
            cleaned_object[k.decode('utf-8')] = v[0].decode('utf-8')
        return cleaned_object

    def _flatten_headers(self, obj):
        flat_data = {}
        for k, v in obj.items():
            flat_data['headers_{}'.format(k)] = v
        return flat_data

    def store_response(self, spider, request, response):
        data = {
            'status': response.status,
            'domain': get_domain(response.url),
            'url': response.url,
            'html': str(response.body).lstrip("b'").strip("'")
                .replace("\\n", "")
                .replace("\\t", "")
                .replace("\\\\", "\\"),
            'created': datetime.now()
        }
        data.update(self._flatten_headers(self._clean_headers(response.headers)))
        self.WebLink(meta={'id': get_urn(response.url)}, **data).save()

    def _read_data(self, spider, request):
        try:
            return self.WebLink.get(id=get_urn(request.url)).to_dict()
        except Exception as e:
            return None

    def _request_key(self, request):
        return to_bytes(request_fingerprint(request))



import uuid
from scrapy.utils.log import configure_logging
from twisted.internet import reactor
from scrapy import signals
from scrapy.crawler import Crawler, CrawlerRunner
from scrapy.settings import Settings
from invana_bot.utils.job import generate_job_id
from invana_bot.settings.default import DEFAULT_SETTINGS


class InvanaBotJobGeneratorBase(object):
    """


    """
    runner = CrawlerRunner()

    def __init__(self,
                 job_id=None,
                 settings=None,
                 **kwargs):

        if settings is None:
            raise Exception("settings should be set")
        self.settings = settings
        self.job_id = self.set_job_id(job_id=job_id)
        self.set_logger()

    @staticmethod
    def set_job_id(job_id=None):
        if job_id is None:
            job_id = generate_job_id()
        return job_id

    @staticmethod
    def set_logger():
        configure_logging({'LOG_FORMAT': '%(levelname)s: %(message)s'})

    @staticmethod
    def _validate_urls(urls):
        if type(urls) is None:
            raise Exception("urls should be list type.")
        if len(urls) is 0:
            raise Exception("urls length should be at least one.")

    def get_settings(self):
        return self.settings

    def start_job(self, job=None, callback_fn=None):
        print(job)
        spider_job = job['spider_job']
        runner = job['runner']
        spider_cls = spider_job['spider_cls']
        spider_settings = spider_job['spider_settings']
        spider_kwargs = spider_job['spider_kwargs']

        def engine_stopped_callback():
            runner.transform_and_index(callback_fn=callback_fn)

        if callback_fn:
            print("""
==========================================================
WARNING: callback_fn is {}
==========================================================
Since start_job is called with callback_fn, make sure you end the reactor if you want the spider process to
stop after the callback function is executed. By default callback_fn=None will close the reactor.

To write a custom callback_fn

def callback_fn():
    print ("Write your own callback logic")
    from twisted.internet import reactor
    reactor.stop()
==========================================================
        """.format(callback_fn))

        spider = Crawler(spider_cls, Settings(spider_settings))
        spider.signals.connect(engine_stopped_callback, signals.engine_stopped)
        self.runner.crawl(spider, **spider_kwargs)
        """
        d = runner.crawl(spider, **spider_kwargs)
        # d.addBoth(engine_stopped_callback)
        """
        reactor.run()



#!/usr/bin/env python

from setuptools import setup, find_packages, __version__ as setuptools_version


setup(name='invana-bot',
      version='0.1.34',
      description='A web spider framework that can transform websites into datasets'
                  ' with Crawl, Transform and Index workflow.',
      author='Ravi Raja Merugu',
      author_email='ravi@invanalabs.ai',
      url='https://github.com/invanalabs/invana-bot',
      packages=find_packages(
          exclude=("bin", "dist", "docs", "example", "tests", "dist")
      ),
      install_requires=[
          'elasticsearch==6.1.1',
          'feedparser==5.2.1',
          'lxml==4.1.1',
          'pymongo==3.7.2',
          'requests==2.21.0',
          'Scrapy==1.6.0',
          'pyyaml==5.1',
          'python-slugify '
      ],
      entry_points={
          'console_scripts': ['invana-bot = invana_bot.cmd.run:invana_bot_run']
      },
      )



from invana_bot.spiders.web import InvanaBotSingleWebCrawler
from scrapy.spiders import Rule
from invana_bot.utils.config import validate_spider_config
from scrapy.linkextractors import LinkExtractor
from .base import RunnerEngineBase


class InvanaBotSingleCrawlerRunnerEngine(RunnerEngineBase):
    """

    InvanaBotSingleCrawlerRunnerEngine


      spider_id: blog_list
      allowed_domains:
        - "blog.scrapinghub.com"
      extractors:
      - extractor_type: ParagraphsExtractor
        extractor_id: paragraphs
        data_storage:
          include_url: true
          collection_name: paragraphs
      - extractor_type: MetaTagExtractor
        extractor_id: meta_tag
        data_storage:
          include_url: true
      settings:
        data_storages:
        - storage_id: default
          storage_type: mongodb
          connection_uri: mongodb://127.0.0.1/spiders_data_index
          collection_name: blog_list
        bot_name: InvanaBot


    """

    def __init__(self,
                 job_id=None,
                 spider_config=None,
                 manifest=None,
                 spider_cls=None,
                 context=None,
                 # manifest_validator=None,
                 **kwargs
                 ):
        """ansf

        :param spider_config: single spider data
        :param manifest: over all manifest data, including spiders, transformations
        :param job_id: this is used for transformating the data of a single job(job_id).
        :param spider_cls:
        :param context: extra data that you may want to access in parse() of spider.
        :param extra_arguments: extra data that you may want to access in parse() of spider.
        # :param manifest_validator: Validator for spider_cls
        """

        self.spider_config = spider_config
        self.manifest = manifest
        self.job_id = job_id
        self.spider_cls = spider_cls
        self.context = context or {}
        self.extra_arguments = context or {}
        # self.manifest_validator = manifest_validator

    def crawl(self):
        errors = validate_spider_config(self.manifest)
        if len(errors) == 0:
            cti_job = self.run()
            return cti_job, errors
        else:
            return None, errors

    def validate_pipe(self):
        must_have_keys = ["spider_id", "extractors"]
        optional_keys = ["traversals"]
        for key in must_have_keys:
            if key not in self.manifest.keys():
                raise Exception(
                    "invalid parser data, should have the following keys; {}".format(",".join(must_have_keys)))

    def validate_traversal(self):
        pass  # TODO - implement this

    def validate_extractor(self):
        pass  # TODO - implement this

    def get_traversals(self):
        return self.manifest.get("traversals", [])

    def get_extractors(self):
        return self.manifest.get("extractors", [])

    def generate_spider_kwargs(self):
        extractor = LinkExtractor()
        rules = [
            Rule(extractor, follow=True)  # TODO - add regex types of needed.
        ]
        print(self.manifest)
        spider_kwargs = {
            "start_urls": self.spider_config['start_urls'],
            "allowed_domains": [],
            "rules": rules,
            "spider_config": self.spider_config,
            "manifest": self.manifest,
            "context": self.context,
            # "default_storage":
        }
        spider_kwargs.update(self.extra_arguments)
        return spider_kwargs

    def run(self):
        spider_cls = self.spider_cls or InvanaBotSingleWebCrawler
        spider_kwargs = self.generate_spider_kwargs()
        return {"spider_cls": spider_cls, "spider_kwargs": spider_kwargs}
